SystemOrganization addCategory: #ClassesUseDiscover!
SystemOrganization addCategory: #'ClassesUseDiscover-Tests'!

!ClassDescription methodsFor: '*ClassesUseDiscover' stamp: 'di 2/17/2000 22:17'!
induceMDFault
	"Stache a copy of the methodDict in the organization slot (hack!!),
	and set the methodDict to nil.  This will induce an MD fault on any message send.
	See: ClassDescription>>recoverFromMDFault
	and ImageSegment>>discoverActiveClasses."

	organization := Array with: methodDict with: organization.
	methodDict := nil.
	self flushCache! !

!ClassDescription methodsFor: '*ClassesUseDiscover' stamp: 'MarianoMartinezPeck 1/11/2010 10:47'!
recoverFromMDFault
	"This method handles methodDict faults to support, eg, discoverActiveClasses (qv)."
	methodDict == nil ifFalse: [^ self].
	"The organization slot should be an array with both, organization and methodDict. This is done in the method induceMDFault"
	(organization isMemberOf: Array) ifFalse: [^ self error: 'Error recovering from MD fault. If the fault was induced using the message induceMDFault, then the organization slot should be an Array with both, organization and methodDict'].
	methodDict := organization first.
	organization := organization second.
! !

!ClassDescription methodsFor: '*ClassesUseDiscover' stamp: 'Mariano 1/4/2010 16:40'!
recoverFromMDFaultWithTrace
	"This method handles emthodDict faults to support, eg, discoverActiveClasses (qv)."
	methodDict == nil ifFalse: [^ self].
	self recoverFromMDFault.
	self environment at: #MDFaultDict ifPresent:
		[:faultDict | faultDict at: self name put:
			(String streamContents:
				[:strm | (thisContext stackOfSize: 20) do: [:item | strm print: item; cr]])]

"Execute the following statement to induce MD fault tracing.  This means that, not only will all active classes be recorded but, after a test run, MDFaultDict will contain, for every class used, a stack trace showing how it came to be used.  This statement should be executed just prior to any such text, in order to clear the traces.

	Smalltalk at: #MDFaultDict put: Dictionary new.

"! !

TestCase subclass: #CUDClassesUseDiscovererTest
	instanceVariableNames: 'activeClasses inactiveClasses state'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClassesUseDiscover-Tests'!

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/6/2010 14:57'!
mdFaultDict

	^ self class environment at: #MDFaultDict! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/11/2010 11:09'!
minNotUsedClasses
	 ^ 100! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/11/2010 11:08'!
minUsedClasses
	^ 50! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/6/2010 15:33'!
testActiveClassesByCategoryDuringDiscovery

	| aDiscoverer inducedClassesNow |
	
	aDiscoverer := CUDClassesUseDiscoverer startDiscovery.
	aDiscoverer stopDiscovery.

	"I assume that for anything minimal you do, you must use at least 10 different categories"
	self assert: aDiscoverer activeClassesByCategoryDuringDiscovery size > 10.
! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/11/2010 11:09'!
testDiscoverActiveClassesDoingNthing
	| aDiscoverer inducedClassesNow |
	
	aDiscoverer := CUDClassesUseDiscoverer startDiscovery.
	"Not all the Classes are induced with a MD fault because the can bring problems, such as Object, Array, etc. I asssume that at least 	I should have induced a MethodDictionary fault in more than a fifth part of the total classes."
	inducedClassesNow := (Smalltalk allClasses select: [:each | (each instVarNamed: 'methodDict') isNil ]).
	self assert: inducedClassesNow size > (Smalltalk allClasses size / 5) asInteger.
	
	"The classes with MD fault right now must be less than the original because to execute the code from that moment to here, some 	classes were recovered."
	self assert: inducedClassesNow size < aDiscoverer amountOfclassesInducedMDFault. 
	aDiscoverer stopDiscovery.
	"After the stop, there should be none MD fault. The classes that were used sohuld be recovered automatically when needed and the 	inactive ones should be reverted in ClassesUseDiscoverer class."
	self assert: (Smalltalk allClasses noneSatisfy: [:each | (each instVarNamed: 'methodDict') isNil ]).
	
	"I assume that for anything minimal you do, you must use at least 50 classes"
	self assert: aDiscoverer activeClassesDuringDiscovery size > self minUsedClasses.
	
	"I assume that for anything minimal you do, you must NOT use at least 100 classes"
	self assert: aDiscoverer inactiveClassesDuringDiscovery size > self minNotUsedClasses.
	
	self assert: aDiscoverer activeClassesDuringDiscovery size + aDiscoverer inactiveClassesDuringDiscovery size > aDiscoverer amountOfclassesInducedMDFault! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/6/2010 15:02'!
testDiscoverActiveClassesWithTrace
	| aDiscoverer inducedClassesNow |
	
	"Just to be sure for the test"
	self class environment removeKey: #MDFaultDict ifAbsent: [].
	
	"There must be a Dictionary in the correct place"
	aDiscoverer := CUDClassesUseDiscoverer startDiscoveryWithTrace.
	self shouldnt: [self mdFaultDict] raise: [Error]. 
	self assert: self mdFaultDict isDictionary. 
	
	aDiscoverer stopDiscovery.
	
	"I assume that for anything minimal you do, you must use at least 50 classes. So, there must be an element per class recovered."
	self assert: self mdFaultDict keys size > 50.
	
	"The activeClassesDuringDiscovery  includes the Classes that were not induced with faults. However, as they were not recovered, they are not in the mdFaultDict."
	self assert: aDiscoverer activeClassesDuringDiscovery size > self mdFaultDict keys size.
	! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/11/2010 11:09'!
testDiscoveryOnBlocks

	| aDiscoverer inducedClassesNow |
	
	aDiscoverer := CUDClassesUseDiscoverer discoverOn: [Transcript show: 'testDiscoveryOnBlocks'].
	"I assume that for anything minimal you do, you must use at least 50 classes"
	self assert: aDiscoverer activeClassesDuringDiscovery size > self minUsedClasses.
	
	"I assume that for anything minimal you do, you must NOT use at least 100 classes"
	self assert: aDiscoverer inactiveClassesDuringDiscovery size > self minNotUsedClasses.
! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/11/2010 11:11'!
testDiscoveryOnBlocksWithTrace

	| aDiscoverer inducedClassesNow |
	
	aDiscoverer := CUDClassesUseDiscoverer discoverWithTraceOn: [Transcript show: 'testDiscoveryOnBlocks'].
	"I assume that for anything minimal you do, you must use at least 50 classes"
	self assert: aDiscoverer activeClassesDuringDiscovery size > self minUsedClasses.
	
	"I assume that for anything minimal you do, you must NOT use at least 100 classes"
	self assert: aDiscoverer inactiveClassesDuringDiscovery size > self minNotUsedClasses.
	
	"The activeClassesDuringDiscovery  includes the Classes that were not induced with faults. However, as they were not recovered, they are not in the mdFaultDict."
	self assert: aDiscoverer activeClassesDuringDiscovery size > self mdFaultDict keys size.
	
! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/6/2010 16:30'!
testSwapOutInactiveClassesDuringDiscovery
	| aDiscoverer inducedClassesNow |
	
	aDiscoverer := CUDClassesUseDiscoverer startDiscovery.
	aDiscoverer stopDiscovery.
	
	aDiscoverer swapOutInactiveClassesDuringDiscovery.
! !

!CUDClassesUseDiscovererTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/6/2010 15:16'!
testWrongStates
	| aDiscoverer inducedClassesNow |
	"The only way to create them is using the class side method startDiscovery"
	self should: [ CUDClassesUseDiscoverer new ] raise: Error.
	
	"You cannot ask the inactive or active classes if you don't first stop the discovery"
	aDiscoverer := CUDClassesUseDiscoverer startDiscovery.
	self should: [ aDiscoverer activeClassesDuringDiscovery ] raise: CUDWrongStateError.
	self should: [ aDiscoverer inactiveClassesDuringDiscovery ] raise: CUDWrongStateError.
	self should: [ aDiscoverer activeClassesByCategoryDuringDiscovery ] raise: CUDWrongStateError.
	
	aDiscoverer stopDiscovery.

	"Once you have started a discovery you cannot started again until you have stop it again"
	aDiscoverer := CUDClassesUseDiscoverer startDiscovery.
	self should: [ aDiscoverer startDiscovery ] raise: CUDWrongStateError.
	aDiscoverer stopDiscovery.
	
	"Once you have stoped a discovery you cannot stop again until you have start it again"
	aDiscoverer := CUDClassesUseDiscoverer startDiscovery.
	aDiscoverer stopDiscovery.
	self should: [ aDiscoverer stopDiscovery ] raise: CUDWrongStateError.
		! !

TestCase subclass: #CUDDiscoveryNotStartedStateTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClassesUseDiscover-Tests'!

!CUDDiscoveryNotStartedStateTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/5/2010 15:36'!
testInclusionsAndRejections
	| state |
	state := 	CUDDiscoveryNotStartedState new.
	
	"They are in the array of excluded"	
	self deny: (state shouldBeInducedMDFault: Object).
	self deny: (state shouldBeInducedMDFault: MethodDictionary).
	
	"It enherits from Array, which is included in the array of excluded"
	self deny: (state shouldBeInducedMDFault: ArrayedCollection).
	
	self assert: (state shouldBeInducedMDFault: Date).
	self assert: (state shouldBeInducedMDFault: Integer).
		


	
! !

!Behavior methodsFor: '*ClassesUseDiscover' stamp: 'MarianoMartinezPeck 1/6/2010 17:44'!
manageMDFault

	self recoverFromMDFaultWithTrace! !

Error subclass: #CUDWrongStateError
	instanceVariableNames: 'stateName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClassesUseDiscover'!

!CUDWrongStateError class methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/5/2010 15:44'!
forState: aStateName
	^ self new
		initializeWith: aStateName! !

!CUDWrongStateError methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 1/5/2010 15:18'!
initializeWith: aStateName
	stateName := aStateName ! !

!CUDWrongStateError methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 1/5/2010 15:18'!
stateName
	^ stateName! !

Object subclass: #CUDClassesUseDiscoverer
	instanceVariableNames: 'activeClasses inactiveClasses amountOfclassesInducedMDFault state'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClassesUseDiscover'!

!CUDClassesUseDiscoverer commentStamp: 'MarianoMartinezPeck 1/11/2010 09:43' prior: 0!
A CUDClassesUseDiscoverer is a little class which can be used to discovering used and unused classes during a period of time. If you want to build minimal images, it is useful also to detect for example, the amount of classes that are used just to do a simple message invokating. It helps also when some classes are being used and you don't know why.

Little examples of how to use it:

	| aDiscoverer |
	
	aDiscoverer := CUDClassesUseDiscoverer startDiscovery.
	......DO SOMETHING HERE......
	aDiscoverer stopDiscovery.

After that, you can:

	aDiscoverer activeClassesDuringDiscovery.
	aDiscoverer inactiveClassesDuringDiscovery.
	aDiscoverer activeClassesByCategoryDuringDiscovery.
	aDiscoverer swapOutInactiveClassesDuringDiscovery.
	
For more details, see the tests in class CUDClassesUseDiscovererTest.
	
Explanation:

The Virtual Machine pays attention not to crash totally when you put a nil in a Method Category. In addition, there is a possibility to force a MethodDictionary fault (message induceMDFault) with the message. This method copies the methodDict in the organization slot. Actually, it creates an Array with the methodDict and the organization and set it in the organization slot. Finally, it sets a nil to the methodDict instance variable. After this is done, when that class is used and the methodDict is asked (it is very important to acces the methodDict using the accessor) it checks whether it is nil or not. If it is nil, it will search for that methodDict in the Array of the organizaction it will then copy it to the methodDict instance variable and will log in Smalltalk at: #MDFaultDict. Sometimes you will see that you have classes that were used but you do not know why or who used it. To solve this question, we can have that log. But in order to have that trace, you have to use the message startDiscoveryWithTrace instead of startDiscovery. The log will be then put in that dictionary and you can inspect it after the discovery.

Taking that explanation in mind, you can follow these steps:

1)  Send the message CUDClassesUseDiscoverer startDiscovery or startDiscoveryWithTrace which forces a MethodDictionary fault in all classes. Yes!! It sends the message induceMDFault to all classes except classes like Array, Object, Class, Message and MethodDictionary. After this, all classes but those ones, will have the methodDict with nil. 
2) Do a few things, use the system for a while. Maybe even save and resume the image.

3) Stop the discovery. This will stop everything and restore all remaining MethodDictionary faults. It will also process which classes were active and which ones were not.

4) Now, it is easy. All classes that still have a nil in the mehodDict is because they were not used, and those who have something different from nil is because they were. With the message activeClassesDuringDiscovery or inactiveClassesDuringDiscovery you can  and return the active or unactive classes respectively. There is also the message  swapOutInactiveClassesDuringDiscovery which makes up segments by grouping unused classes by system category and write them to disk. There are some categories that are excluded because they have problems, like System and Kernel. 


If you started the discovery with the message startDiscoveryWithTrace, after doing all those steps you can then inspect Smalltalk at: #MDFaultDict and you will see that in the dictionary you will have an element per class. That element is a ByteString that contains all the stack trace. This shows you all the messages and why that class was used. 

This approach of swapping out inactive classes have two big problems:

- The granularity is too high: the class. As soon as only one method is invoked, the whole class is recovered. 
- It is very common to recover all classes doing a simple action. For example, searching for implementors or senders of a certain method. Because of this, when you are developing in your Smalltalk image is very likely that all classes are recovered in less than 5 minutes of work. It is not useful. Browsing or editing code seems not to be a problem, thus. Maybe this approach has more sense in production images rather than development. 

!

!CUDClassesUseDiscoverer class methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/11/2010 11:06'!
discoverOn: aBlock
	| newInstance |
	newInstance := self basicNew 
		initialize.
	newInstance startDiscovery.
	aBlock value.
	newInstance stopDiscovery.
	^ newInstance! !

!CUDClassesUseDiscoverer class methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/11/2010 11:10'!
discoverWithTraceOn: aBlock
	| newInstance |
	newInstance := self basicNew 
		initialize.
	newInstance startDiscoveryWithTrace.
	aBlock value.
	newInstance stopDiscovery.
	^ newInstance! !

!CUDClassesUseDiscoverer class methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/5/2010 15:41'!
new
	self error: 'You have to use the class side method startDiscovery'! !

!CUDClassesUseDiscoverer class methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/5/2010 15:43'!
startDiscovery
	
	^ self basicNew 
		initialize;
		startDiscovery! !

!CUDClassesUseDiscoverer class methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/5/2010 16:08'!
startDiscoveryWithTrace
	
	^ self basicNew 
		initialize;
		startDiscoveryWithTrace! !

!CUDClassesUseDiscoverer methodsFor: 'accessing' stamp: 'Mariano 1/4/2010 14:37'!
activeClasses   
	^ activeClasses ! !

!CUDClassesUseDiscoverer methodsFor: 'accessing' stamp: 'Mariano 1/4/2010 14:22'!
activeClasses: anObject
	activeClasses := anObject! !

!CUDClassesUseDiscoverer methodsFor: 'results' stamp: 'MarianoMartinezPeck 1/6/2010 15:15'!
activeClassesByCategoryDuringDiscovery   
	"Return a dictionary of active classes by system category. The key of each element is a category and the value is a list of the classes of that category that were used.
	Useful for finding kernel categories to minimize swapping."

	| active dict cat list |
	active := self activeClassesDuringDiscovery.
	dict := Dictionary new.
	active do:
		[:each | cat := each category.
		list := dict at: cat ifAbsent: [Array new].
		dict at: cat put: (list copyWith: each)].
	^ dict
! !

!CUDClassesUseDiscoverer methodsFor: 'results' stamp: 'Mariano 1/4/2010 14:38'!
activeClassesDuringDiscovery   
	^ state activeClasses ! !

!CUDClassesUseDiscoverer methodsFor: 'accessing' stamp: 'Mariano 1/4/2010 15:07'!
amountOfclassesInducedMDFault
	^ amountOfclassesInducedMDFault! !

!CUDClassesUseDiscoverer methodsFor: 'accessing' stamp: 'Mariano 1/4/2010 15:07'!
amountOfclassesInducedMDFault: anObject
	amountOfclassesInducedMDFault := anObject! !

!CUDClassesUseDiscoverer methodsFor: 'states' stamp: 'Mariano 1/4/2010 17:19'!
changeToState: aDiscoveryState
	self log: 'I will change from state ', state name, ' to state ', aDiscoveryState name.
	state := aDiscoveryState. 

	! !

!CUDClassesUseDiscoverer methodsFor: 'private' stamp: 'MarianoMartinezPeck 1/6/2010 16:33'!
createAndWriteSegmentForCategoryRoots: roots name: categoryName

	(ImageSegment new
                       copyFromRoots: roots
                       sizeHint: 0
                       areUnique: true)
                               extract;
                               writeToFile: categoryName 
! !

!CUDClassesUseDiscoverer methodsFor: 'accessing' stamp: 'Mariano 1/4/2010 10:44'!
inactiveClasses   
	^ inactiveClasses ! !

!CUDClassesUseDiscoverer methodsFor: 'accessing' stamp: 'Mariano 1/4/2010 14:22'!
inactiveClasses: anObject
	inactiveClasses := anObject! !

!CUDClassesUseDiscoverer methodsFor: 'results' stamp: 'Mariano 1/4/2010 14:38'!
inactiveClassesDuringDiscovery   
	^ state inactiveClasses ! !

!CUDClassesUseDiscoverer methodsFor: 'initialization' stamp: 'Mariano 1/4/2010 11:03'!
initialize
	super initialize.
	self initializeActiveAndInactiveClasses.
	self initializeState.! !

!CUDClassesUseDiscoverer methodsFor: 'initialization' stamp: 'Mariano 1/4/2010 10:15'!
initializeActiveAndInactiveClasses
	
	activeClasses := OrderedCollection new.
	inactiveClasses := OrderedCollection new. ! !

!CUDClassesUseDiscoverer methodsFor: 'initialization' stamp: 'MarianoMartinezPeck 1/5/2010 15:36'!
initializeState
	state := CUDDiscoveryNotStartedState forDiscoverer: self.! !

!CUDClassesUseDiscoverer methodsFor: 'logging' stamp: 'Mariano 1/4/2010 16:23'!
log: aString

	Transcript show: aString; cr.! !

!CUDClassesUseDiscoverer methodsFor: 'private' stamp: 'MarianoMartinezPeck 6/29/2010 16:36'!
prefixOfCategoriesToExclude

	^ #( 'Sys' 'Multilingual' 'Kerne' 'ClassesUseDiscov' 'Collections-Arrayed' 'Graphics-Display Objects')
! !

!CUDClassesUseDiscoverer methodsFor: 'private' stamp: 'MarianoMartinezPeck 1/6/2010 16:26'!
shouldCategoryGoToSegment: aCategory

	^ self prefixOfCategoriesToExclude noneSatisfy: [:each | aCategory beginsWith: each ]! !

!CUDClassesUseDiscoverer methodsFor: 'actions' stamp: 'Mariano 1/4/2010 11:44'!
startDiscovery

	state startDiscovery.
	! !

!CUDClassesUseDiscoverer methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/5/2010 16:11'!
startDiscoveryWithTrace 
	self class environment at: #MDFaultDict put: Dictionary new.
	state startDiscovery.
	! !

!CUDClassesUseDiscoverer methodsFor: 'actions' stamp: 'Mariano 1/4/2010 14:15'!
stopDiscovery

	state stopDiscovery.! !

!CUDClassesUseDiscoverer methodsFor: 'swapping' stamp: 'MarianoMartinezPeck 6/29/2010 15:02'!
swapOutInactiveClassesDuringDiscovery
	"Make up segments by grouping unused classes by system category. It will create a segment per category and it will wrriten to disk with the name of the category."

	| unused groups i roots symbols |
	
	"I need to hold the symbols so that not to get duplicates."
	symbols := Symbol allSymbols.
	
	unused := self inactiveClassesDuringDiscovery.	
	groups := Dictionary new.

	SystemOrganization categories do:
			[ :cat | groups at: cat put: ((SystemOrganization superclassOrder: cat) select: [ :c | unused includes: c ]) asArray ].
		
	groups keys do:
			[ :cat |
			"Not all the classes can be swap out because they bring problems" 
			(self shouldCategoryGoToSegment: cat) ifTrue:
					[ roots := groups at: cat.
					"Reject compact classes as they cannot be roots"
					roots := roots reject: [ :each | each indexIfCompact > 0 ].   
					"For each class I put in a segment, I also have to put its metaclass"
					roots := roots , (roots collect: #class).
					self createAndWriteSegmentForCategoryRoots: roots name: cat ] ].
			
	Smalltalk garbageCollect! !

!CUDClassesUseDiscoverer methodsFor: 'swapping' stamp: 'MarianoMartinezPeck 7/5/2010 14:28'!
swapOutInactiveClassesDuringDiscoveryAllTogether
	"Make up segments by grouping unused classes by system category. It will create a segment per category and it will wrriten to disk with the name of the category."

	| unused groups i roots symbols totalRoots |
	
	"I need to hold the symbols so that not to get duplicates."
	symbols := Symbol allSymbols.
	
	unused := self inactiveClassesDuringDiscovery.	
	groups := Dictionary new.
	totalRoots := Array new.

	SystemOrganization categories do:
			[ :cat | groups at: cat put: ((SystemOrganization superclassOrder: cat) select: [ :c | unused includes: c ]) asArray ].
		
	groups keys do:
			[ :cat |
			"Not all the classes can be swap out because they bring problems" 
			(self shouldCategoryGoToSegment: cat) ifTrue:
					[ roots := groups at: cat.
					"Reject compact classes as they cannot be roots"
					roots := roots reject: [ :each | each indexIfCompact > 0 ].   
					"For each class I put in a segment, I also have to put its metaclass"
					totalRoots := totalRoots, roots , (roots collect: #class).  ] ].
			
	self createAndWriteSegmentForCategoryRoots: roots name: 'aTest'.	
		
	Smalltalk garbageCollect! !

!CUDClassesUseDiscoverer methodsFor: 'swapping' stamp: 'MarianoMartinezPeck 7/5/2010 15:27'!
swapOutInactiveClassesDuringDiscoveryClassPerClass
	"Make up segments by grouping unused classes by system category. It will create a segment per category and it will wrriten to disk with the name of the category."

	| unused groups i roots symbols totalRoots |
	
	"I need to hold the symbols so that not to get duplicates."
	symbols := Symbol allSymbols.

	unused := self inactiveClassesDuringDiscovery.	
	"Don't swap compact classes as they cannot be root in ImageSegment"	
	unused do: [: aClass | ((aClass indexIfCompact = 0) and: [self shouldCategoryGoToSegment: aClass category] ) 
		ifTrue: [ self createAndWriteSegmentForCategoryRoots: (Array with: aClass with: aClass theMetaClass) name: aClass name. ]].		
		
	Smalltalk garbageCollect! !

Object subclass: #CUDDiscoveryState
	instanceVariableNames: 'classesUseDiscover'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClassesUseDiscover'!

CUDDiscoveryState subclass: #CUDDiscoveryFinishedState
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClassesUseDiscover'!

!CUDDiscoveryFinishedState methodsFor: 'results' stamp: 'MarianoMartinezPeck 1/5/2010 15:10'!
activeClasses
	self log: 'Amount of active classes during discovery: ', classesUseDiscover activeClasses size asString.
	^ classesUseDiscover activeClasses! !

!CUDDiscoveryFinishedState methodsFor: 'results' stamp: 'MarianoMartinezPeck 1/5/2010 15:10'!
inactiveClasses
	self log: 'Amount of inactive classes during discovery: ', classesUseDiscover inactiveClasses size asString.
	^ classesUseDiscover inactiveClasses ! !

CUDDiscoveryState subclass: #CUDDiscoveryNotStartedState
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClassesUseDiscover'!

!CUDDiscoveryNotStartedState class methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 7/5/2010 16:13'!
arrayOfClassesNotToInduceMDFault
	"There are several classes that for different reasons, cannot be swapped out."
	| allClasses coreClasses protoObjectSubclasses classesSpecialArray veryUsedClasses |
	
	"These classes because they are too core and used by the code to swapout itseld"
	coreClasses :=  {Array. Object. Class. Message. MethodDictionary. CUDClassesUseDiscoverer. ProtoObject. ImageSegment. ImageSegmentRootStub.} .
	
	"Subclasses of ProtoObject because they do not implement some methods that Object does and that are used by the swapping stuff."
	protoObjectSubclasses := ProtoObject subclasses.
	
	"These classes cannot be swapped because they are used by the VM"
	classesSpecialArray := Smalltalk specialObjectsArray select: [ :each | each isBehavior ].
	
	"These classes cannot be swapped because they install the segment (they are used) even before the segment can be written into a file"
	veryUsedClasses := { Cursor. CursorWithMask. EncodedCharSet.}.
	
	allClasses := coreClasses, protoObjectSubclasses, classesSpecialArray, veryUsedClasses.
	^ allClasses.
	
	! !

!CUDDiscoveryNotStartedState methodsFor: 'private' stamp: 'MarianoMartinezPeck 1/6/2010 17:24'!
induceMDFaultToAllClasses
	| count |
	count := 0.
	
	self log: 'I will induce all - almost - classes'.
	Smalltalk allClasses select: [:each | self shouldBeInducedMDFault: each ] thenDo:
	 	[:each | each induceMDFault.
		count := count + 1 ].
	
	self log: 'Amount of classes induced: ', count asString.
	classesUseDiscover amountOfclassesInducedMDFault: count

	! !

!CUDDiscoveryNotStartedState methodsFor: 'private' stamp: 'MarianoMartinezPeck 6/29/2010 16:23'!
shouldBeInducedMDFault: aClass
	^ self class arrayOfClassesNotToInduceMDFault noneSatisfy: [:eachRootClass | (aClass == eachRootClass or: [ eachRootClass inheritsFrom: aClass ]) ].
	! !

!CUDDiscoveryNotStartedState methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/5/2010 15:36'!
startDiscovery
	
	self log: 'Starting with the discovery'.
	self induceMDFaultToAllClasses.
	self changeToState: (CUDDiscoveryRunningState forDiscoverer: classesUseDiscover).

	! !

CUDDiscoveryState subclass: #CUDDiscoveryRunningState
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ClassesUseDiscover'!

!CUDDiscoveryRunningState methodsFor: 'private' stamp: 'Mariano 1/4/2010 17:21'!
recoverInactiveClassedFromMDFault
"Restore all remaining MD faults and return the active classes"
	self log: 'Recover of inactive classes'.
	self classesUseDiscover inactiveClasses do: [ :c | c recoverFromMDFault ]! !

!CUDDiscoveryRunningState methodsFor: 'private' stamp: 'Mariano 1/4/2010 17:19'!
setsToDiscovererActiveAndInactiveClasses
	"Restore all remaining MD faults and return the active classes"

	| inactive active |
	inactive := OrderedCollection new.
	active := OrderedCollection new.
	Smalltalk allClasses do:
			[ :each | (each instVarNamed: 'methodDict') 
							ifNil: [ inactive addLast: each ] 
							ifNotNil: [ active addLast: each ] ].
	classesUseDiscover inactiveClasses: inactive.
	classesUseDiscover activeClasses: active.
	
	self log: 'Classes set to discoverer.'.
	self log: 'Amount of inactive classes: ', inactive size asString.
	self log: 'Amount of active classes: ', active size asString.
	! !

!CUDDiscoveryRunningState methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/5/2010 15:35'!
stopDiscovery
	self log: 'Stopping the discovery'.
	self setsToDiscovererActiveAndInactiveClasses.
	self recoverInactiveClassedFromMDFault.
	self changeToState: (CUDDiscoveryFinishedState forDiscoverer: classesUseDiscover).! !

!CUDDiscoveryState class methodsFor: 'as yet unclassified' stamp: 'Mariano 1/4/2010 15:32'!
forDiscoverer: aClassesUseDiscover
	^ self new
		classesUseDiscover: aClassesUseDiscover! !

!CUDDiscoveryState methodsFor: 'results' stamp: 'MarianoMartinezPeck 1/5/2010 15:11'!
activeClasses
	self signalWrongStateError.! !

!CUDDiscoveryState methodsFor: 'private - state' stamp: 'Mariano 1/4/2010 14:07'!
changeToState: aDiscoveryState

	classesUseDiscover changeToState: aDiscoveryState.
! !

!CUDDiscoveryState methodsFor: 'accessing' stamp: 'Mariano 1/4/2010 11:48'!
classesUseDiscover
	^ classesUseDiscover! !

!CUDDiscoveryState methodsFor: 'accessing' stamp: 'Mariano 1/4/2010 11:48'!
classesUseDiscover: anObject
	classesUseDiscover := anObject! !

!CUDDiscoveryState methodsFor: 'results' stamp: 'MarianoMartinezPeck 1/5/2010 15:24'!
inactiveClasses
	self signalWrongStateError.! !

!CUDDiscoveryState methodsFor: 'logging' stamp: 'Mariano 1/4/2010 15:57'!
log: aString

	Transcript show: aString; cr.! !

!CUDDiscoveryState methodsFor: 'error handling' stamp: 'MarianoMartinezPeck 1/5/2010 15:36'!
signalWrongStateError
	(CUDWrongStateError forState: self class name asString) signal! !

!CUDDiscoveryState methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/5/2010 15:24'!
startDiscovery
	self signalWrongStateError.! !

!CUDDiscoveryState methodsFor: 'actions' stamp: 'MarianoMartinezPeck 1/5/2010 15:24'!
stopDiscovery
	self signalWrongStateError.! !
